/*
 * Maintain an /etc/hosts - alike file, with the addition that a counter
 * is increased for each entry whenever it is accessed. The file format is
 * ip-addr - name counter timestamp
 * Each machine may have only one single name and no aliases. The counter
 * counts references to that ip address and the timestamp is used for
 * expiring addresses. The counter must be exactly 10 digits long, with
 * zeroes as prefix if the counter is less than 1 billion :-)
 * Entries in /etc/hosts format are accepted (with aliases and without
 * counter), but then the counter cannot be maintained.
 * The file format can easily be created, modified etc. with awk scripts.
 *
 * Oh yes, I almost forgot -- the file MUST BE SORTED BY IP ADDRESS. So
 * it's maybe not such a good idea to start with /etc/hosts after all.
 * Remember that you can NOT sort the file by IP-Addresses with sort(1).
 * It does alpha compares. We need numerical compares.
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include "hosts.h"

/*
 * We use a little table for faster lookups. You can choose its size
 * in bits. The table will have 2^bits entries and a size of 2^bits*4
 */

#define HTTABLEBITS	10
#define HTTABLESIZE	(1<<HTTABLEBITS)

/*
 * max length of host name
 */

#define HTFQNMAXLEN	63

static long httable[HTTABLESIZE];

/*
 * static file pointer to open file
 */

static char  htfname[1024];
static FILE *htfile = NULL;

/*
 * open file and init table
 */

int
HTOpenFile (char *file)
{
  char tentry[512];
  unsigned long ipaddr, thisval, oldval=0;
  long oldpos=0, thispos;
  unsigned int ade[4];

#ifdef DEBUG3
  fprintf (stderr, "HTOpenFile()\n");
#endif

  if (htfile != NULL)
    HTCloseFile ();

  if ((htfile = fopen (file, "r+")) == NULL) {
#ifdef DEBUG
    fprintf (stderr, "HTOpenFile(%s) failed.\n", file);
#endif
    return 1;
  }

  /*
   * remember file name
   */
  strcpy (htfname, file);

  while (!feof (htfile)) {
    thispos = ftell (htfile);
    if (fgets (tentry, 512, htfile) == NULL)
      break;
    if (!isdigit (tentry[0]))
      continue;
    if (sscanf (tentry, "%u.%u.%u.%u", 
		&(ade[0]), &(ade[1]), &(ade[2]), &ade[3]) != 4)
      continue;

    ipaddr  = (ade[0]<<24) | (ade[1]<<16) | (ade[2]<<8) | ade[3];
    thisval = (ipaddr >> (32 - HTTABLEBITS));

    if ((thisval != oldval) && thisval!=0) {
      for (oldval++; oldval<thisval; oldval++)
	httable[oldval-1] = thispos;
      
#ifdef DEBUG3
      fprintf (stderr, "HTOpenFile: Entry %8lu Pos %8l\n",
	       thisval-1, thispos);
#endif

      httable[thisval-1] = thispos;
      oldpos = thispos;
      oldval = thisval;
    }
  }
  /*
   * init rest of table
   */

  for (oldval++, thispos=ftell(htfile); oldval<HTTABLESIZE; oldval++)
    httable[oldval-1] = thispos;

  /*
   * leave file open and return OK
   */
  return 0;
}

/*
 * Look up an IP number in network byte order. If found, the FQN is
 * copied to the given buffer. If it is longer than 64 characters,
 * the 63 _last_ characters are copied.
 * Returns NULL if entry doesn't exist, buffer otherwise
 */

char *
HTLookupEntry (unsigned long address, char *buffer)
{
  static char tentry[512], *ptr, *ptr2;
  unsigned long ipaddr, theval;
  long thispos, minpos, maxpos;
  unsigned int ade[4];
  int counter;

  if (htfile == NULL)
    return NULL;

#ifdef DEBUG3
  fprintf (stderr, "HTLookupEntry (%u.%u.%u.%u)\n",
	   address>>24, (address>>16)&255, (address>>8)&255, address&255);
#endif

  theval = (address >> (32 - HTTABLEBITS));
  minpos = (theval == 0) ? 0 : httable[theval-1];
  maxpos = httable[theval];

#ifdef DEBUG3
  fprintf (stderr, "   minpos %8ld maxpos %8ld    diff %8ld\n",
	   minpos, maxpos, maxpos-minpos);
#endif

  if (maxpos - minpos <= 0) {
#ifdef DEBUG2
    fprintf (stderr, "HTLookupEntry (%u.%u.%u.%u) failed\n",
	     address>>24, (address>>16)&255, (address>>8)&255, address&255);
#endif
    return NULL;
  }

  /*
   * try an infile binary search.
   */

  while (42) {
    fseek (htfile, (maxpos+minpos)/2, SEEK_SET);
    /*
     * search next newline
     */

    if (fgets (tentry, 512, htfile) == NULL)
      break;

    /*
     * read entry. skip entries that aren't ip addresses.
     */

    while (42) {
      thispos = ftell (htfile);

#ifdef DEBUG3
      fprintf (stderr, "   minpos %8ld maxpos %8ld thispos %8ld searching\n",
	       minpos, maxpos, thispos);
#endif

      if (thispos >= maxpos) {
#ifdef DEBUG3
	fprintf (stderr, "   minpos %8ld maxpos %8ld thispos %8ld breakout\n",
		 minpos, maxpos, thispos);
#endif
	break;
      }

      if (fgets (tentry, 512, htfile) == NULL)
	break;

      if (!isdigit (tentry[0]))
	continue;
      
      if (sscanf (tentry, "%u.%u.%u.%u", 
		  &(ade[0]), &(ade[1]), &(ade[2]), &ade[3]) == 4)
	break;
    }

    if (feof (htfile) || thispos >= maxpos)
      break;

    ipaddr = (ade[0]<<24) | (ade[1]<<16) | (ade[2]<<8) | ade[3];

#ifdef DEBUG3
    fprintf (stderr, "   compare %u.%u.%u.%u\n",
	     ade[0], ade[1], ade[2], ade[3]);
#endif

    if (address == ipaddr) { /* found */
      break;
    }
    else if (ipaddr > address) {
#ifdef DEBUG3
      fprintf (stderr, "   minpos %8ld maxpos %8ld  newmax %8ld narrowing\n",
	       minpos, maxpos, thispos);
#endif
      maxpos = thispos;
    }
    else /* if (ipaddr < address) */ {
#ifdef DEBUG3
      fprintf (stderr, "   minpos %8ld maxpos %8ld  newmin %8ld narrowing\n",
	       minpos, maxpos, thispos);
#endif
      minpos = thispos;
    }
  }

  /*
   * if we haven't found it so far, do a linear search. maybe the
   * binary search got stuck somewhere inbetween. because of this
   * binary narrowing, the area in question should be pretty small
   */

  if (address != ipaddr) {
    fseek (htfile, minpos, SEEK_SET);

    while (!feof (htfile)) {
      thispos = ftell (htfile);

#ifdef DEBUG3
      fprintf (stderr, "   minpos %8ld maxpos %8ld thispos %8ld lin search\n",
	       minpos, maxpos, thispos);
#endif

      if (thispos >= maxpos)
	break;

      if (fgets (tentry, 512, htfile) == NULL)
	break;

      if (!isdigit (tentry[0]))
	continue;
      
      if (sscanf (tentry, "%u.%u.%u.%u", 
		  &(ade[0]), &(ade[1]), &(ade[2]), &ade[3]) != 4)
	continue;
    
      ipaddr = (ade[0]<<24) | (ade[1]<<16) | (ade[2]<<8) | ade[3];

#ifdef DEBUG3
    fprintf (stderr, "   compare %u.%u.%u.%u\n",
	     ade[0], ade[1], ade[2], ade[3]);
#endif

      if (address <= ipaddr)
	break;
    }
  }

  if (address != ipaddr) {
    /*
     * not found
     */
#ifdef DEBUG2
    fprintf (stderr, "HTLookupEntry (%u.%u.%u.%u) failed\n",
	     address>>24, (address>>16)&255, (address>>8)&255, address&255);
#endif
    return NULL;
  }

  /*
   * address found. extract name and increase counter
   */

#ifdef DEBUG3
  fprintf (stderr, "HTLookupEntry() found\n");
#endif

  ptr = tentry;

  while (*ptr && !isspace(*ptr))
    ptr++;
  while (isspace(*ptr))
    ptr++;

  if (*ptr != '-') {
    /*
     * entry is in general /etc/hosts format, *ptr is name
     */
    ptr2 = ptr;
    while (*ptr2 && !isspace (*ptr2))
      ptr2++;
    if (ptr2 - ptr > HTFQNMAXLEN) {
      /*
       * oops, too long
       */
      memcpy (buffer, ptr2-HTFQNMAXLEN, HTFQNMAXLEN);
      buffer[HTFQNMAXLEN] = '\0';
    }
    else {
      *ptr2 = '\0';
      strcpy (buffer, ptr);
    }
#ifdef DEBUG3
    fprintf (stderr, "   conventional entry name '%s'\n",
	     buffer);
#endif

  }
  else {
    /*
     * entry is like we want to have it
     */
    ptr++;
    while (isspace (*ptr))
      ptr++;
    
    ptr2 = ptr;
    while (*ptr2 && !isspace (*ptr2))
      ptr2++;
    if (ptr2 - ptr > HTFQNMAXLEN) {
      /*
       * oops, too long
       */
      memcpy (buffer, ptr2-HTFQNMAXLEN, HTFQNMAXLEN);
      buffer[HTFQNMAXLEN] = '\0';
    }
    else {
      if (*ptr2)
	*ptr2++ = '\0';
      strcpy (buffer, ptr);
    }
    while (*ptr2 && !isdigit (*ptr2))
      ptr2++;

    if (isdigit (*ptr2)) {
      counter = atoi (ptr2);

      /*
       * fseek to the position of the counter and write new counter value
       */
      fseek   (htfile, thispos+(ptr2-tentry), SEEK_SET);
      fprintf (htfile, "%010d", counter+1);
      /*
       * must fflush() before we can read again
       */
      fflush  (htfile);
    }
#ifdef DEBUG3
    fprintf (stderr, "   newstyle entry counter %d\n", counter+1);
#endif
  }

  /*
   * wow, done
   */

#ifdef DEBUG2
  fprintf (stderr, "HTLookupEntry (%u.%u.%u.%u) success %s\n",
	   address>>24, (address>>16)&255, (address>>8)&255, address&255,
	   buffer);
#endif

  return buffer;
}

/*
 * Update our beloved file. We merge the file with new data into a new file
 * and then replace the original file. We interface with a source of new
 * entries with a callback function. This function is expected to deliver
 * new entries ordered by IP address. It will be called with an increasing
 * counter as parameter. This counter starts at 0 and is increased by 1 each
 * time (it doesn't jump and doesn't run backwards). The function returns
 * non-zero if there are no more entries. We also rebuild the index table
 * anew.
 */

int
HTAddEntries (int (*callback) (int, unsigned long *, char *, int *))
{
  char tentry[512], newname[512], savename[512], *tname;
  unsigned long ipaddr, thisval, newaddr, oldval=0;
  int index, counter=0, newflag, ctr=0;
  long thispos, oldpos=0;
  unsigned int ade[4];
  FILE *tfile;
  time_t now;

  /*
   * open temp file
   */

#ifdef DEBUG3
  fprintf (stderr, "HTAddEntries()\n");
#endif

  if ((tname = tempnam ("/tmp", "_w3o_")) == NULL) {
#ifdef DEBUG
    fprintf (stderr, "HTAddEntries () tempnam failed\n");
#endif
    return 1;
  }
  if ((tfile = fopen (tname, "w")) == NULL) {
#ifdef DEBUG
    fprintf (stderr, "HTAddEntries couldn't open temp file %s\n",
	     tname);
#endif
    free (tname);
    return 2;
  }

  /*
   * get first new entry
   */

  newflag = (*callback) (counter++, &newaddr, newname, &ctr);

  if (newflag) {
#ifdef DEBUG
    fprintf (stderr, "HTAddEntries () no new data\n");
#endif
    /*
     * hey, there's nothing new. there's really no reason to merge
     * anything
     */
    fclose (tfile);
    unlink (tname);
    free   (tname);
    return 0;
  }

  /*
   * rewind source file
   */

  fseek (htfile, 0, SEEK_SET);

  /*
   * time stamp
   */

  now = time (NULL);

  /*
   * start merging
   */

  while (!feof (htfile)) {
    /*
     * search next entry
     */

    if (fgets (tentry, 512, htfile) == NULL)
      break;
    if (!isdigit (tentry[0])) {
      fprintf (tfile, "%s", tentry);
      continue;
    }
    if (sscanf (tentry, "%u.%u.%u.%u",
		&(ade[0]), &(ade[1]), &(ade[2]), &ade[3]) != 4) {
      fprintf (tfile, "%s", tentry);
      continue;
    }

    ipaddr  = (ade[0]<<24) | (ade[1]<<16) | (ade[2]<<8) | ade[3];

    /*
     * if the entry exists, we ignore the new addition
     */

    while (newflag==0 && newaddr == ipaddr) {
#ifdef DEBUG3
      fprintf (stderr, "   ignoring %u.%u.%u.%u %s\n",
	       (unsigned int) ((newaddr >> 24) & 255),
	       (unsigned int) ((newaddr >> 16) & 255),
	       (unsigned int) ((newaddr >>  8) & 255),
	       (unsigned int) ((newaddr      ) & 255),
	       newname);
#endif
      /*
       * read new entry
       */
      newflag = (*callback) (counter++, &newaddr, newname, &ctr);
    }

    while (newflag==0 && newaddr<ipaddr) {
      /*
       * add new entry
       */

      thispos = ftell (tfile);

#ifdef DEBUG3
      fprintf (stderr, "   adding %u.%u.%u.%u %s\n",
	       (unsigned int) ((newaddr >> 24) & 255),
	       (unsigned int) ((newaddr >> 16) & 255),
	       (unsigned int) ((newaddr >>  8) & 255),
	       (unsigned int) ((newaddr      ) & 255),
	       newname);
#endif

      fprintf (tfile, "%u.%u.%u.%u\t-\t%s\t\t%010d\t%lu\n",
	       (unsigned int) ((newaddr >> 24) & 255),
	       (unsigned int) ((newaddr >> 16) & 255),
	       (unsigned int) ((newaddr >>  8) & 255),
	       (unsigned int) ((newaddr      ) & 255),
	       newname,
	       ctr?ctr:1,
	       (unsigned long) now);
      /*
       * perhaps a new table entry?
       */
      thisval = (newaddr >> (32 - HTTABLEBITS));

      if ((thisval != oldval) && thisval != 0) {
	for (oldval++; oldval<thisval; oldval++)
	  httable[oldval-1] = thispos;

	httable[thisval-1] = thispos;
	oldpos = thispos;
	oldval = thisval;
      }
      /*
       * read new entry
       */
      newflag = (*callback) (counter++, &newaddr, newname, &ctr);
    }

    /*
     * copy from source
     */

    thispos = ftell (tfile);
    fprintf (tfile, "%s", tentry);

    thisval = (ipaddr >> (32 - HTTABLEBITS));

    if ((thisval != oldval) && thisval != 0) {
      for (oldval++; oldval<thisval; oldval++)
	httable[oldval-1] = thispos;

      httable[thisval-1] = thispos;
      oldpos = thispos;
      oldval = thisval;
    }
  }

  /*
   * there may be some new entries left
   */

  while (newflag==0) {
    /*
     * add new entry
     */
    
    thispos = ftell (tfile);

#ifdef DEBUG3
    fprintf (stderr, "   adding %u.%u.%u.%u %s\n",
	     (unsigned int) ((newaddr >> 24) & 255),
	     (unsigned int) ((newaddr >> 16) & 255),
	     (unsigned int) ((newaddr >>  8) & 255),
	     (unsigned int) ((newaddr      ) & 255),
	     newname);
#endif
    
    fprintf (tfile, "%u.%u.%u.%u\t-\t%s\t\t%010d\t%lu\n",
	     (unsigned int) ((newaddr >> 24) & 255),
	     (unsigned int) ((newaddr >> 16) & 255),
	     (unsigned int) ((newaddr >>  8) & 255),
	     (unsigned int) ((newaddr      ) & 255),
	     newname,
	     ctr?ctr:1,
	     (unsigned long) now);
    /*
     * perhaps a new table entry?
     */
    thisval = (newaddr >> (32 - HTTABLEBITS));
    
    if ((thisval != oldval) && thisval != 0) {
      for (oldval++; oldval<thisval; oldval++)
	httable[oldval-1] = thispos;
      
      httable[thisval-1] = thispos;
      oldpos = thispos;
      oldval = thisval;
    }
    /*
     * read new entry
     */
    newflag = (*callback) (counter++, &newaddr, newname, &ctr);
  }
  /*
   * init rest of table
   */

  for (oldval++, thispos=ftell(htfile); oldval<HTTABLESIZE; oldval++)
    httable[oldval-1] = thispos;

  /*
   * close files, rename temp file to original and reopen it
   */

  fclose (tfile);
  fclose (htfile);

  /*
   * now we unlink() the source file. the original file will continue
   * to exist for all other running instances
   */

  unlink (htfname);

  /*
   * now we copy the temp file to the real one. we first copy it to
   * a different file names, as we don't want new instances to open
   * the unfinished file
   */

  sprintf (savename, "%s.new", htfname);

  if ((tfile = fopen (tname, "r")) == NULL) {
#ifdef DEBUG
    fprintf (stderr, "HTAddEntries () panic couldn't reopen temp file %s\n",
	     tname);
#endif
    free (tname);
    return 1;
  }
  if ((htfile = fopen (savename, "w")) == NULL) {
#ifdef DEBUG
    fprintf (stderr, "HTAddEntries () panic couldn't open new file %s\n",
	     savename);
#endif
    fclose (tfile);
    free (tname);
  }

  /*
   * copy
   */

  while (!feof (tfile)) {
    if ((index = fread (tentry, 1, 512, tfile)) == 0)
      break;
    fwrite (tentry, 1, index, htfile);
  }

  fclose (tfile);
  fclose (htfile);

  /*
   * okay, now rename the new file to the real one
   */

  if (rename (savename, htfname)) {
#ifdef DEBUG
    fprintf (stderr, "HTAddEntries () rename (%s,%s) failed errno %d\n",
	     savename, htfname, errno);
#endif
    unlink (savename);
    free (tname);
    return 1;
  }

  /*
   * everything went fine so far. now we can delete the temp file safely
   */

  if (unlink (tname)) {
#ifdef DEBUG
    fprintf (stderr, "HTAddEntries () couldn't unlink temp file %s\n",
	     tname);
#endif
    /* error ignored */
  }

  /*
   * reopen the new file
   */

  if ((htfile = fopen (htfname, "r+")) == NULL) {
#ifdef DEBUG
    fprintf (stderr, "HTAddEntries () oops could not reopen %s\n",
	     htfname);
#endif
  }

#ifdef DEBUG3
  fprintf (stderr, "HTAddEntries() end\n");
#endif

  return 0;
}

/*
 * close file and destroy table. there's nothing to destroy yet
 */

void
HTCloseFile (void)
{
  if (htfile) {
#ifdef DEBUG3
    fprintf (stderr, "HTCloseFile(%s)\n", htfname);
#endif
    fclose (htfile);
    htfile = NULL;
  }
  return;
}

/*
 * test code

int
TestCallBack (int c, unsigned long *addr, char *name)
{
  if (c == 0) {
    *addr = (141L<<24) | (2L<<16) | (18L<<8) | 99;
    strcpy (name, "rose.fpp.de");
    return 0;
  }
  else if (c == 1) {
    *addr = (141L<<24) | (3L<<16) | (18L<<8) | 99;
    strcpy (name, "rose.here.elsewhere");
    return 0;
  }
  else
    return 1;
}

void
main (int argc, char *argv[])
{
  unsigned long ipaddr;
  unsigned int ade[4];
  char buffer[64];

  if (HTOpenFile (argv[1]))
    return;
  if (sscanf (argv[2], "%u.%u.%u.%u", 
	      &(ade[0]), &(ade[1]), &(ade[2]), &(ade[3])) != 4) {
    fprintf (stderr, "illegal ip number\n");
    HTCloseFile ();
    return;
  }
  ipaddr = (ade[0]<<24) | (ade[1]<<16) | (ade[2]<<8) | ade[3];

  if (HTLookupEntry (ipaddr, buffer)) {
    fprintf (stdout, "host found: '%s'\n", buffer);
  }
  else {
    fprintf (stdout, "host not found\n");
  }

  HTAddEntries (TestCallBack);

  HTCloseFile ();
}

 *
 */
